open Semantics

exception Not_applicable

let update_output = ref (fun _ -> failwith "unset" : int * Strtree.t -> unit)
let current_taglen = ref max_int

let diff currattr nextattr =
  begin
    match currattr.b, nextattr.b with
    | false, true -> ["B", (fun attr -> { attr with b = true })]
    | true, false -> raise Not_applicable
    | _ -> []
  end @ begin
    match currattr.ems, nextattr.ems with
    | No, S -> ["S", (fun attr -> { attr with ems = S })]
    | No, Em -> ["EM", (fun attr -> { attr with ems = Em })]
    | Em, No -> ["EM", (fun attr -> { attr with ems = No })]
    | Em, S -> ["S", (fun attr -> { attr with ems = S })]
    | S, No | S, Em -> raise Not_applicable
    | _ -> []
  end @ begin
    match currattr.i, nextattr.i with
    | false, true -> ["I", (fun attr -> { attr with i = true })]
    | true, false -> raise Not_applicable
    | _ -> []
  end @ begin
    match currattr.tt, nextattr.tt with
    | false, true -> ["TT", (fun attr -> { attr with tt = true })]
    | true, false -> raise Not_applicable
    | _ -> []
  end @ begin
    let dif = nextattr.u - currattr.u in
    let utag = ("U", (fun attr -> { attr with u = attr.u + 1 })) in
    if dif < -3 then assert false else
    if dif < 0 then raise Not_applicable
    else if dif = 0 then []
    else if dif = 1 then [utag]
    else if dif = 2 then [utag; utag]
    else if dif = 3 then [utag; utag; utag]
    else assert false
  end @ begin
    match currattr.size, nextattr.size with
    | None, None -> []
    | Some x, Some y when x = y -> []
    | Some _, None -> raise Not_applicable
    | _, Some y -> [string_of_int y,
		    (fun attr -> { attr with size = Some y })]
  end @ begin
    match currattr.color, nextattr.color with
    | None, None -> []
    | Some x, Some y when x = y -> []
    | Some _, None -> raise Not_applicable
    | _, Some y -> [Printf.sprintf "%c" y,
		    (fun attr -> { attr with color = Some y })]
  end

let rec split1 = function
  | [] -> []
  | hd :: tl ->
      (hd, tl) ::
      List.map
	(fun (x, y) -> (x, hd :: y))
	(split1 tl)

let rec permutate f acc = function
    (* [XXX] optimization for mutiple "U" tags *)
  | [] -> f acc
  | l ->
      List.iter
	(fun (x, y) -> permutate f (x :: acc) y)
	(split1 l)

(*
let hashcons tagname attr stack =
  (tagname, attr, 0) :: stack
*)

let nullhash = Hashtbl.hash []
let stackhash = function
  | [] -> nullhash
  | (_, _, h) :: _ -> h
let hashcons tagname attr stack =
  let h = Hashtbl.hash (tagname, attr, stackhash stack) in
  (tagname, attr, h) :: stack

module H =
  struct
    type t = (string * attr * int) list (* stack *) * int (* pos *)
    let equal = (=)
    let hash (stack, pos) = Hashtbl.hash (stackhash stack, pos)
  end

module StackHash = Hashtbl.Make(H)

let hashtbl = StackHash.create 1000000
let _ = StackHash.clear hashtbl

let memo stack pos new_taglen =
  try
    let old_taglen = StackHash.find hashtbl (stack, pos) in
    if old_taglen > new_taglen then raise Not_found else
    true
  with Not_found -> begin
    StackHash.replace hashtbl (stack, pos) new_taglen;
    false
  end

let rec try_pop_and_pl_and_push
    (acc : Strtree.t)
    (taglen : int)
      (* invariant: taglen is the length of tags in acc *)
    (stack : (string * attr * int) list)
    (stack_taglen : int)
      (* invariant: stack_taglen is the length of tags necessary for emptying stack *)
    (currattr : attr)
    (nextattr : attr)
    (str : string)
    (pos : int)
    (rest : Semantics.t list) =
  if taglen + stack_taglen >= !current_taglen then begin
    if Debug.level >= 2 then Format.eprintf "cut: %s@." (Strtree.to_string acc)
  end else
    match stack with
    | [] ->
	begin
	  try 
	    try_push acc taglen [] 0 default_attr nextattr str pos rest
	  with Not_applicable -> ()
	end
    | (tag, prevattr, _) :: rst ->
	let new_acc = Strtree.append acc (Printf.sprintf "</%s>" tag) in
	let new_taglen = taglen + String.length tag + 3 in
	try_pop_and_pl_and_push new_acc new_taglen rst (stack_taglen - String.length tag - 3) prevattr nextattr str pos rest;
	try_pl_and_push acc taglen stack stack_taglen currattr nextattr str pos rest;

and try_pl_and_push acc taglen stack stack_taglen currattr nextattr str pos rest =
  if taglen + stack_taglen >= !current_taglen then begin
    if Debug.level >= 2 then Format.eprintf "cut: %s@." (Strtree.to_string acc)
  end else
    try
      try_push acc taglen stack stack_taglen currattr nextattr str pos rest;
    with Not_applicable -> begin
      try
	try_push
	  (Strtree.append acc "<PL>")
	  (taglen + 4)
	  (hashcons "PL" currattr stack)
	  (stack_taglen + 5)
	  { currattr with u = 0; ems = No; i = false; b = false; tt = false }
	  nextattr
	  str
	  pos
	  rest
      with Not_applicable -> ()
    end

and try_push acc taglen stack stack_taglen currattr nextattr str pos rest =
  if taglen + stack_taglen >= !current_taglen then begin
    if Debug.level >= 2 then Format.eprintf "cut: %s@." (Strtree.to_string acc)
  end else
    let applicable_tags = diff currattr nextattr in
    permutate
      (fun x ->
	let new_acc, new_taglen, new_stack, new_stack_taglen, new_state = 
	  List.fold_left
	    (fun (acc, taglen, stack, stack_taglen, state) (tagname, update) -> 
	      Strtree.append acc (Printf.sprintf "<%s>" tagname),
	      taglen + String.length tagname + 2,
	      hashcons tagname state stack,
	      stack_taglen + String.length tagname + 3,
	      update state)
	    (acc, taglen, stack, stack_taglen, currattr)
	    x
	in iter (Strtree.append new_acc str) new_taglen new_stack new_stack_taglen new_state pos rest)
      []
      applicable_tags

and iter acc taglen stack stack_taglen currattr pos rest =
  if taglen + stack_taglen >= !current_taglen then begin
    if Debug.level >= 2 then Format.eprintf "cut: %s@." (Strtree.to_string acc)
  end else
    if memo stack pos taglen then begin
      if Debug.level >= 2 then Format.eprintf "hit: %s@." (Strtree.to_string acc)
    end else
      match rest with
      | [] ->
	  !update_output 
	    (List.fold_left
	       (fun (taglen, acc) (tagname, _, _) ->
		 taglen + String.length tagname + 3,
		 Strtree.append acc (Printf.sprintf "</%s>" tagname))
	       (taglen, acc)
	       stack)
      | Text(nextattr, str) :: rest ->
	  try_pop_and_pl_and_push acc taglen stack stack_taglen currattr nextattr str (pos + 1) rest
      | Space(spcattr, str) :: rest ->
          (* [XXX] we can do better *)
          (* [XXX] we could merge with the attributes of the next Text *)
	  let nextattr = un_spc_ctx currattr spcattr in
	  try_pop_and_pl_and_push acc taglen stack stack_taglen currattr nextattr str (pos + 1) rest

let f sem =
  iter
    Strtree.create (* acc *)
    0 (* taglen *)
    [] (* stack *)
    0 (* stack_taglen *)
    default_attr (* currattr *)
    0 (* pos *)
    sem (* rest *)
